#!/usr/bin/perl -w

use Tie::File;
use Fcntl 'O_RDONLY';
use Getopt::Long;
use File::Spec;
use File::Basename;
use strict;
use warnings;
use Scalar::Util qw(looks_like_number);
use List::Util qw(max min sum);

$| = 1;

# Reads a phenotype file of tab seperated values and performs checks on
# them, printing out a report at the end detailing any errors found.
#
#
#
#
#
# usage perl phenotype_qc.pl -f filename [-v]
#
# V3.0 changed to use tied arrays rather than load data set as hash into memory
# v3.1 Added Sample fingerprint

# get today's date
my $time = localtime;

my $delimiter = '\t';
my $filepath;
my $today;
my $name;
my $path;
my $suffix;
my @suffixlist = (".txt", ".tab", ".tsv");
my $file;
my $verbose_flag = 0;
my $help = 0;
my $missing = 'NA';
my $headerfile = '';
my $idcolumn = 0; #default column for sample ids
my $column;       #column for a single phenotype  
my %columns;      #column for multiple phenotypes
my $columnfile;   #file to provide multiple phenotypes for %columns
my %headers;

GetOptions
 (
 "f|file=s"      => \$filepath,      #file to be checked
 "v|verbose"     => \$verbose_flag,  #specify verbose output (prints to logfile)
 "l|lookup=s",   => \$headerfile,    #specify a header file mapping in case they are not human readable
 "i|idcolumn=s", => \$idcolumn,      #specify the id containing column, if it's not the first one
 "c|column=s"    => \$column,        #check a single phenotype in the file
 "s|columnfile=s"=> \$columnfile,    #specify a file containing many phenotypes to check
 "m|missing=s"   => \$missing,       #specify the missing value identifier, if not NA
 "h|help"        => \$help           #print the help message
 );


if ($help)
 {
 usage();
 exit;
 }

if (!$filepath)
 {
 usage();
 exit;
 }
 
 
if ($time=~/\w+?\s(\w+?)\s+(\d+)\s.+?\s(\d+)/)
 {
 $today = sprintf("%02d", $2).$1.sprintf("%4d", $3);
 }

if ($headerfile)
 {
 %headers = get_headers($headerfile);
 }

if ($columnfile)
 {
 %columns = get_columns($columnfile);
 }
elsif ($column)
 {
 $columns{$column} = 1;
 }
else
 {
 %columns = all_columns($filepath);
 }


($name, $path, $suffix) = fileparse($filepath,@suffixlist);

my $outputfilename = 'Summaryv1.'.$name.'_'.$today.'.txt';
my $logfilename = 'Logv1.'.$name.'.'.$today.'.txt';

# these files are appended to, check at start if they exist, if so rename
my $fpfile = 'FPv1.'.$name.'.'.$today.'.txt';   #fingerprint of the phenotypes (columns)
my $fpsfile = 'FPSv1.'.$name.'.'.$today.'.txt'; #fingerprint of the samples (rows)
$fpfile = check_file($fpfile);
$fpsfile = check_file($fpsfile);
$outputfilename = check_file($outputfilename);
$logfilename = check_file($logfilename);

open OUTPUT, ">$outputfilename" or die "Unable to open $outputfilename\n";
open LOGFILE, ">$logfilename" or die "unable to open logfile\n";
open STDERR, ">error.log" or die "unable to open error log file\n";

print OUTPUT "\t\tPhenotype QC report\n\n";
print OUTPUT "\tInput file:\t\t\t$filepath\n";
print OUTPUT "\tDate created:\t\t\t$today\n";
print OUTPUT "\tMissing value identifier:\t$missing\n\n";
#print OUTPUT "\tCreated William Rayner 2016 version 3.0\n";
#print OUTPUT "Values for each item are only shown if there are less than 50 items\n";
print OUTPUT "The verbose flag (set using -v on the command line) will result in \n";
print OUTPUT "all values will be written to the log file\n";

if ($verbose_flag)
 {
 print OUTPUT "Verbose flag is set\n\n";
 }
print OUTPUT "------------------------------------------------------------------\n\n";

get_data($filepath, $idcolumn, $missing, $fpfile, \%columns);
my %column_numbers = get_column_numbers($filepath, \%columns);
$column_numbers{$idcolumn} = 1;
get_fps($filepath, $idcolumn, $missing, $fpsfile, \%column_numbers);


sub get_column_numbers
 {
 my $file = $_[0];
 my %cols = %{$_[1]};
 my %coln;
 
 open IN, "$file" or die $!;
 my $headerline = <IN>;
 close IN;
 chomp $headerline;
 my @temp = split(/\t/, $headerline); 
 
 for (my $i = 0; $i <= $#temp; $i++)
  {
  if ($cols{$temp[$i]} or $cols{$headers{$temp[$i]}})
   {
   $coln{$i} = 1;
   }
  else
   {
   $coln{$i} = 0;
   }
  }
  
 return %coln;
 }

sub all_columns
 {
 my $file = $_[0];
 my %allCols;
 open IN, "$file" or die $!;
 my $header = <IN>;
 close IN;
 chomp $header;
 my @columns = split(/\t/, $header, -1);
 $allCols{$_}++ for (@columns);
 return %allCols;
 }

sub get_columns
 {
 my $file = $_[0];
 my %cols;
 open IN, "$file" or die $!;
 while (<IN>)
  {
  chomp;
  my @temp = split/\t/;
  $cols{$temp[0]} = 1;
  }
 return %cols
 }

sub check_file
 {
 my $filename = $_[0];
 my $newfilename = $filename;
 if (-e $filename)
  {
  if ($filename =~ /(.*?v)(\d+)(\..*)$/)
   {
   my $num = $2;
   $num++;
   $newfilename = $1.$num.$3;
   $newfilename = check_file($newfilename);
   }
  else
   {
   print "ERROR: parsing filename failed\n";
   }
  }
 else
  {
  $newfilename = $filename;
  }
 return $newfilename;
 }

sub get_fps
 {
 my $filename = $_[0];
 my $sampleidcolumn = $_[1];
 my $missing = $_[2];
 my $fpsfile = $_[3];
 my %cols = %{$_[4]};
 
 
 print "Reading $filename\n";
 open IN, "$filename";
 while (<IN>)
  {
  my @data;
  chomp;
  my @temp = split/\t/;
  for (my $i = 0; $i <= $#temp; $i++)
   {
   if ($cols{$i})
    {
    push (@data, $temp[$i]);
    }
   }
  my $type = get_type(\@data, $missing, $fpsfile);
  }
 close IN; 
 } 

sub get_data
 {
 my $filename = $_[0];
 my $sampleidcolumn = $_[1];
 my $missing = $_[2];
 my $fpfile = $_[3];
 my %columns = %{$_[4]};
 
 print "Reading $filename\n";
 open IN, "$filename";
 my $line = <IN>;
 close IN;
 my @elements = split(/\t/, $line, -1);
 my @array;
 
 tie @array, 'Tie::File', $filename, recsep => "\t", memory => 200_000_000, mode => O_RDONLY or die;
  
 my $records = @array;
 my $columns = @elements;
 
 #print "$records\n";
 #print "$columns\n";
 
 my @data2;
 
 for (my $j = 0; $j < $#elements; $j++)
  {
  
  if ($columns{$elements[$j]} or $columns{$headers{$elements[$j]}})
   {
   my @data;
  
   my $k = 0;
   #print "Processing $array[$j]\n";
   for (my $i = $j; $i < $records; $i += $#elements)
    {
    my $result = index($array[$i], "\n");     #see where carriage return is
    if ($result != -1)                        #carriage return found in the string
     {
     my $resultlength = $result + 1;          #add 1 to the position to get overall length
     my $stringlength = length($array[$i]);   #see how long string is
     if ($stringlength == $resultlength)      #carriage return is the last one
      {
      my $val = $array[$i];                   #make a copy so not attempting file modification, opened read only
      $val =~ s/\n//;
      $data2[$k-1] = $val;
      $data2[$k-1] =~ s/\"//g;
      $data2[$k-1] =~ s/^\s+$/$missing/;
      }
     else
      {
      my @temp = split(/\n/, $array[$i]);
      $data[$k] = $temp[1];
      $data[$k] =~ s/\"//g;
      $data[$k] =~ s/^\s+$/$missing/;    
      $data2[$k-1] = $temp[0];
      $data2[$k-1] =~ s/\"//g;
      $data2[$k-1] =~ s/^\s+$/$missing/; 
      }
     }
    else                                      #no carriage return assign as normal
     {
     $data[$k] = $array[$i];
     $data[$k] =~ s/\"//g;
     $data[$k] =~ s/^\s+$/$missing/;          #set to missing any entry comprising all spaces
     }
    $k++;
    }
  
   if ($j != $sampleidcolumn) #don't try and summarise the sample ids idcolumn
    {
    #print "$j $sampleidcolumn\n";
    process_data(\@data, $missing, $fpfile);
    }
   else
    {
    print OUTPUT "\nSkipping Sample id idcolumn: $array[$j]\n";
    print "Skipping $array[$j] as idcolumn containing sample id\n";
    }
   
   } 
  else
   {
   if ($headerfile)
    {
    if ($headers{$elements[$j]})
     {
     print "Skipping $headers{$elements[$j]}\n";
     }
    else
     {
     print "Skipping $elements[$j]\n";
     }
    }
   else
    {
    print "Skipping $elements[$j]\n";
    }
   }
  }
  
  if (@data2)
   {
   process_data(\@data2, $missing, $fpfile);
   }
  
 untie @array; 
 }

sub process_data
 {
 my @data = @{$_[0]};
 my $missing = $_[1];
 my $fpfile = $_[2];
 
 if ($headerfile)
  {
  if ($headers{$data[0]})
   {
   print "Processing $headers{$data[0]}";
   }
  else
   {
   print "Processing $data[0]";
   }
  }
 else
  {
  print "Processing $data[0]";
  }

 #determine the idcolumn type (numeric, text, coded) 
 my $type = get_type(\@data, $missing, $fpfile);
 print ". Type $type\n";
 
 if ($headerfile)
  {  
  if ($headers{$data[0]})
   {
   print OUTPUT "\nExamining $headers{$data[0]},";    
   }
  else
   {
   print OUTPUT "\nExamining $data[0],"; 
   }
  }
 else
  {
  print OUTPUT "\nExamining $data[0],"; 
  } 
 if ($type =~ /.*CODED$/)
  {
  if ($type eq 'NUMERIC CODED')
   {
   print OUTPUT  " it appears to be a numeric coded set of values\n";
   coded(\@data, $missing);
   }
  elsif ($type eq 'TEXT CODED')
   {
   print OUTPUT  " it appears to be a text coded set of values\n";
   coded(\@data, $missing);
   } 
  elsif ($type eq 'MIXED CODED')
   {
   print OUTPUT  " it appears to be a mixed coded set of values\n";
   coded(\@data, $missing);
   } 
  }
 elsif ($type eq 'NUMERIC')
  {
  print OUTPUT  " it appears to be a numeric set of values\n";
  numeric(\@data, $missing);
  }
 elsif ($type eq 'TEXT')
  {
  print OUTPUT  " it appears to be a set of text values\n";
  coded(\@data, $missing);
  }
 elsif ($type eq 'MIXED')
  {
  print OUTPUT  " it appears to be a mixed set of values\n";
  coded(\@data, $missing);
  }
 elsif ($type eq 'MISSING')
  {
  print OUTPUT  " it appears to consist entirely of missing values ($missing)\n";
  }
 else
  {
  
  }
 }

sub coded
 {
 my @data = @{$_[0]};
 my $missingval = $_[1];
 my $missingcount = 0;
 
 my @datavalues = splice(@data, 0, 1); 
 my $total = @data;
 my %count;
 
 print OUTPUT "Found $total values\n";

 for (my $i = 0; $i <= $#data; $i++)
  {
  if ($data[$i] eq $missingval)
   {
   $missingcount++;
   }
  else
   {
   $count{$data[$i]}++;
   }
  }
 my @codes = keys %count;
 my $numcodes = @codes;
 
 if ($numcodes == 1)
  {
  print OUTPUT "With $numcodes value:\n";
  }
 else
  {
  print OUTPUT "With $numcodes different values:\n";
  }
 
 if ($missingcount)
  {
  print OUTPUT "Found $missingcount missing values ($missingval)\n"; 
  }
 
 if ($numcodes < 50 and $numcodes != 0)
  {
  print OUTPUT "Values found are:\n\tCode\tcount\n";
 
  foreach my $code (@codes)
   {
   print OUTPUT "\t$code\t$count{$code}\n";
   }
  } 
 }

sub numeric
 {
 my @rawdata = @{$_[0]};
 my $missingval = $_[1];
 my $j = 0;
 my @data;
 my @datavalues = splice(@rawdata, 0, 1); 

 # remove missing values
 for (my $i = 0; $i <= $#rawdata; $i++)
  {
  $rawdata[$i] =~ s/\"//g;
  if ($rawdata[$i])
   {
   if ($rawdata[$i] ne $missing)
    {
    $data[$j] = $rawdata[$i];
    $j++;
    }
   }
  }
 my $total = @rawdata; 
 my $missingcount = $total - $j;
   
 my $average = 0;
 
 my $maxval = max @data;
 my $minval = min @data;
 $total = @data;
 my $sum = sum @data;
 my $median = median(\@data);
 $median = sprintf("%0.3g", $median); 
 
 
 if ($total != 0)
  {
  $average = sprintf("%0.3g",$sum/$total);
  }

 my $stddev = std_dev(\@data, $average); 
 $stddev = sprintf("%0.3g",$stddev);
 
 print OUTPUT "Found $total numeric values\n";
 print OUTPUT "Found $missingcount missing values ($missingval)\n";
 print OUTPUT "Mean:\t\t\t$average\n";
 print OUTPUT "Max Value:\t\t$maxval\n";
 print OUTPUT "Min Value:\t\t$minval\n";
 print OUTPUT "Std Deviation:\t\t$stddev\n";
 print OUTPUT "Median:\t\t\t$median\n";
 print OUTPUT "Checking for outliers +/-3 S.D.\n";
 if ($verbose_flag)
  {
  print LOGFILE "\tVALUE\tGREATER OR LESS THAN 3 SDs\n";  
  } 
 print_outliers(\@data, $average, $stddev, $verbose_flag);
 
 }

sub print_outliers
 {
 my @data = @{$_[0]};
 my $mean = $_[1];
 my $stddev = $_[2];
 my $verbose_flag = $_[3];
 my $limit = 3; # number of std deviations from mean
 my $over = 0;
 my $under = 0;
 my $greater = $mean+($stddev*3);
 my $less = $mean-($stddev*3);
 #print "Greater $greater\nLess $less\n";
 my $sdflag = 0;
  
 for (my $i = 0; $i <= $#data; $i++)
  {
  if ($data[$i])
   {
   if ($data[$i] >= $greater)
    {
    $over++;
    if ($verbose_flag)
     {
     print LOGFILE "\t$data[$i]\tGREATER\n";
     }
    $sdflag = 1;
    }
   elsif ($data[$i] <= $less)
    {
    if ($verbose_flag)
     {
     print LOGFILE "\t$data[$i]\tLESS\n";
     }
    $under++;
    $sdflag = 1;
    }
   } 
  }
   
 if ($sdflag == 0)
  {
  print OUTPUT "...None\n";
  }
 else
  {
  print OUTPUT ">$limit S.D.s\t\t$over\n<$limit S.D.s\t\t$under\n";
  }
 }

sub median 
 {
 my @data = @{$_[0]};
 my @sorted = sort {$a <=> $b} @data; 
 my $median = 0;
 
 if (($#sorted % 2) == 0) # even number of elements
  {
  # median value is the average of the two middle elements
  my $middle = $#sorted/2;
  $median = ($sorted[$middle]+$sorted[$middle-1])/2;
  }
 else # odd number of elements
  {
  $median = $sorted[($#sorted-1)/2];
  }
 return $median;
 }

sub std_dev 
 {
 my @data = @{$_[0]};
 my $mean = $_[1];
 my $sum_x2 = 0;
 my $var = 0;
 my $numvals = @data; 
 my $count = 0;
 my $stddev;
 
 for (my $i = 0; $i < $numvals; $i++)
  {
  if ($data[$i])
   {
   $sum_x2 += (($data[$i] - $mean) ** 2);
   $count++;
   }
  }
  
 if ($count != 0)
  {
  $var = ( $sum_x2 / $count );
  $stddev = sqrt($var);
  }
 else 
  {
  $var = 0;
  $stddev = 0;
  }
  
 return $stddev;
 }

sub get_type
 {
 my @data = @{$_[0]};
 my $missingval = $_[1];
 my $fpfile = $_[2];
 
 my $num = 0;
 my $notnum = 0;
 my $total = 0;
 my %datahash;
 my $type = 'TEXT';
 my @datavalues = splice(@data, 0, 1); 
 my $miss = 0;
 my $totalkeys = 0;
 my $pct = 0;
 
 open FP, ">>$fpfile" or die $!;
 
 for (my $i = 0; $i < $#data; $i++) # check if the value is numeric, 
  {
  if ($data[$i] and $data[$i] ne $missingval) #ensure value is present and isn't the designated missing value id
   {
   my $isnum = looks_like_number($data[$i]);
   if ($isnum)
    {
    $num++;
    }
   else
    {
    $notnum++; 
    }
   $total++; 
   $datahash{$data[$i]}++;
   }
  else
   {
   $miss++;
   }
  } 
 
 if ($total)
  { 
  if ($num == $total) #all numeric
   {
   $type = 'NUMERIC';
   }
  elsif ($notnum == $total) #all non numeric 
   {
   $type = 'TEXT';
   }
  else #mixed numeric/text
   {
   $type = 'MIXED';
   }
  
  $totalkeys = keys %datahash;
  $pct = $totalkeys/$total*100;
  if ($pct < 0.1 or $totalkeys <= 20) # check as a % of total number of values, with hard coded limits for small cohorts
   {
   $type = $type.' CODED';
   }
  }
 else
  {
  $type = 'MISSING';
  }
  
 #prints out a "fingerprint" of the data array for use in future checks
 if ($headerfile)
  {
  if ($headers{$datavalues[0]})
   {
   print FP "$headers{$datavalues[0]}\t$num\t$notnum\t$total\t$totalkeys\t$miss\t$type\t";
   }
  else
   {
   print FP "$datavalues[0]\t$num\t$notnum\t$total\t$totalkeys\t$miss\t$type\t";
   }
  }
 else
  {
  print FP "$datavalues[0]\t$num\t$notnum\t$total\t$totalkeys\t$miss\t$type\t";
  }
 print FP "\t\t\n";
 close FP;
 return $type;  
 }

sub usage
 {
 print "\nUsage:\nperl phenotype-qc.pl -f <phenotypefile> [-v -m <missingvalue> -i <sampleididcolumn> -l <headerfilemapping> -c <columnname> -s <fileofcolumns>]\n";
 print "\n\n";
 printusage("-f --file", "phenotype file", "Tab delimited phenotype file, all rows should contain the same number of columns");
 printusage("-i --idcolumn", "ID column", "Set the column that contains the sample ids, column numbers start from 0, default is 0");
 printusage("-m --missing", "Missing value", "Set the missing value identifier, default is NA");
 printusage("-c --column", "Column", "Specify a single column to check, if using a header file this can be either the human readable or original header");
 printusage("-s --columnfile", "Column file", "Provide a file containing a list of columns to extract, one per row. As with the single column checking can be either header, takes precedence over -c");
 printusage("-l --linker", "Header file", "Provide an optional header file for the data set, for use where the headers within the file are not human readable");
 printusage("-v --verbose", "", "Optional flag to increase verbosity in the log file");
 printusage("-h --help", "", "Show this help message");
 print "\n\n";
 }

sub getwidth
 {
 my $output = `stty size`;
 my @rowcols = split(/\s/, $output);
 my $cols = $rowcols[1];
 if (!$cols)
  {
  $cols = 80;
  }
 return $cols;
 } 
 
sub printusage
 {
 my $option = $_[0];
 my $file = $_[1];
 my $text = $_[2];
 
 my $cols = getwidth(); 

 printf("%-18s", $option);
 printf("%-15s", $file);
 my $textwidth = $cols - 36; #33 plus 3 spaces
 my $newtext = substr ($text, 0, $textwidth);
 
 if (length($text) > $textwidth)
  {
  printf("  %*s", $textwidth, $newtext);
  for (my $i = $textwidth; $i < length($text); $i+= $textwidth)
   {
   my $newtext = substr ($text, $i, $textwidth);
   $newtext =~ s/^\s//;
   printf("%33s", " ");
   printf("   %*s", -$textwidth, $newtext);
   }
  }
 else
  {
  printf("  %*s", -$textwidth, $newtext);
  }
 print "\n"; 
 } 
  
sub get_headers
 {
 my $file = $_[0];
 my %header;
 
 open IN, "$file" or die $!;
 while (<IN>)
  {
  chomp;
  my @temp = split/\t/;
  $header{$temp[0]} = $temp[1];
  }
 close IN;
 return %header;
 }

